import copy
from collections import namedtuple
from typing import List, Union
from marshmallow import fields
import marshmallow as ma

import importlib_metadata

Import = namedtuple("Import", "import_path,alias")


class DataType:
    model_type = None
    marshmallow_field = None
    schema_type = None
    mapping_type = None
    use_dumper = False

    class ModelSchema(ma.Schema):
        type = fields.String(required=True)

    def __init__(self, definition, key, model, schema):
        self.definition = definition
        self.key = key
        self.model = model
        self.schema = schema

    def _copy_definition(self, **extras):
        ret = copy.deepcopy(self.definition)
        for k, v in extras.items():
            if v is not None:
                ret[k] = v
        return ret

    def model_schema(self, **extras):
        return self._copy_definition(**extras)

    def json_schema(self, **extras):
        return self._copy_definition(type=self.schema_type, **extras)

    def mapping(self, **extras):
        return self._copy_definition(type=self.mapping_type, **extras)

    def marshmallow(self, **extras):
        ret = copy.deepcopy(self.definition.get("marshmallow", {}))
        if self.marshmallow_field:
            ret.setdefault("field-class", self.marshmallow_field)
        ret.setdefault("validators", []).extend(self.marshmallow_validators())
        for k, v in extras.items():
            if v is not None:
                ret[k] = v
        return ret

    def marshmallow_validators(self) -> List[str]:
        return []

    def imports(self, *extra) -> List[Import]:
        return extra

    def facet(self, nested_facet):
        return nested_facet

    def dumper_class(self, data):  # NOSONAR
        return None


class DataTypes:
    def __init__(self) -> None:
        self.datatype_map = {}

    def _prepare_datatypes(self):
        if not self.datatype_map:
            for entry in importlib_metadata.entry_points(
                group="oarepo_model_builder.datatypes"
            ):
                for dt in entry.load():
                    self.datatype_map[dt.model_type] = dt

    def get_datatype(self, data, key, model, schema) -> Union[DataType, None]:
        datatype_class = self.get_datatype_class(data.get("type", None))
        if datatype_class:
            return datatype_class(data, key, model, schema)
        return None

    def get_datatype_class(self, datatype_type):
        self._prepare_datatypes()
        return self.datatype_map.get(datatype_type)


datatypes = DataTypes()



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioProxiesBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_proxies"
    class_config = "proxies-current-resource"
    template = "proxies"



from marshmallow import fields
from .model_validation import model_validator
from marshmallow_union import Union
from marshmallow.exceptions import ValidationError

from .utils import ExtendablePartSchema
import typing


class StrictString(fields.String):
    def _deserialize(self, value, attr, data, **kwargs) -> typing.Any:
        if value is None or isinstance(value, str):
            return super()._deserialize(value, attr, data, **kwargs)
        raise ValidationError(
            f"String value expected, found {type(value)} with value {repr(value)[:30]}..."
        )

    def _serialize(self, value, attr, obj, **kwargs) -> typing.Union[str, None]:
        if value is None or isinstance(value, str):
            return super()._serialize(value, attr, obj, **kwargs)
        raise ValidationError(
            f"String value expected, found {type(value)} with value {repr(value)[:30]}..."
        )


class Property(ExtendablePartSchema):
    facets = fields.Nested(lambda: model_validator.validator_class("property-facets")())
    sample = Union(
        [
            fields.Nested(lambda: model_validator.validator_class("property-sample")()),
            fields.List(fields.Boolean()),  # or array of booleans
            fields.List(fields.Integer()),  # or array of booleans
            fields.List(fields.Float()),  # or array of booleans
            fields.List(fields.String()),  # or array of strings
            StrictString(),  # use this constant value for the sample data
        ],
        required=False,
    )
    jsonschema = fields.Nested(
        lambda: model_validator.validator_class("property-jsonschema")()
    )
    mapping = fields.Nested(
        lambda: model_validator.validator_class("property-mapping")()
    )
    marshmallow = fields.Nested(
        lambda: model_validator.validator_class("property-marshmallow")()
    )
    sortable = fields.Nested(
        lambda: model_validator.validator_class("property-sortable")()
    )


class ObjectProperty(ExtendablePartSchema):
    required = fields.Boolean(required=False)



import json

from oarepo_model_builder.utils.camelcase import camel_case
from oarepo_model_builder.utils.jinja import split_package_name
from oarepo_model_builder.validation import InvalidModelException, model_validator

from .datatypes import DataType
from marshmallow import fields


class ObjectDataType(DataType):
    schema_type = "object"
    mapping_type = "object"
    marshmallow_field = "ma_fields.Nested"
    model_type = "object"

    class ModelSchema(DataType.ModelSchema):
        properties = fields.Nested(
            lambda: model_validator.validator_class("properties", strict=False)()
        )

    def marshmallow(self, **extras):
        ret = super().marshmallow(**extras)

        schema_class = ret.get("schema-class", None)
        if not schema_class:
            schema_class = camel_case(self.key) + "Schema"

        package_name = split_package_name(self.model.record_schema_class)

        schema_class = self._get_class_name(package_name, schema_class)

        fingerprint = json.dumps(
            self.definition, sort_keys=True, default=lambda x: repr(x)
        ).encode("utf-8")

        schema_class = self._find_unique_schema_class(
            self.model.setdefault("known-classes", {}), schema_class, fingerprint
        )

        self.model.known_classes[schema_class] = fingerprint

        ret["schema-class"] = schema_class

        return ret

    def _find_unique_schema_class(self, known_classes, schema_class, fingerprint):
        if schema_class in known_classes:
            # reuse class with the same fingerprint
            if fingerprint != known_classes[schema_class]:
                for i in range(100):
                    candidate = f"{schema_class}_{i}"
                    if candidate not in known_classes:
                        schema_class = candidate
                        break
                    if fingerprint == known_classes[candidate]:
                        schema_class = candidate
                        break
                else:
                    raise InvalidModelException(
                        f"Too many marshmallow classes with name {schema_class}. Please specify your own class names"
                    )

        return schema_class

    def _get_class_name(self, package_name: str, class_name: str):
        if "." not in class_name:
            return f"{package_name}.{class_name}"
        if class_name.startswith("."):
            package_path = package_name.split(".")
            while class_name.startswith("."):
                if package_path:
                    package_path = package_path[:-1]
                class_name = class_name[1:]
            if package_path:
                class_name = f"{'.'.join(package_path)}.{class_name}"
        return class_name


class NestedDataType(ObjectDataType):
    schema_type = "object"
    mapping_type = "nested"
    marshmallow_field = "ma_fields.Nested"
    model_type = "nested"

    def facet(self, nested_facet):
        return f"NestedLabeledFacet(path={self.path}, {nested_facet})"


class FlattenDataType(DataType):
    schema_type = "object"
    mapping_type = "flatten"
    marshmallow_field = "ma_fields.Raw"
    model_type = "flatten"


class ArrayDataType(DataType):
    schema_type = "array"
    mapping_type = None
    marshmallow_field = "ma_fields.List"
    model_type = "array"

    class ModelSchema(DataType.ModelSchema):
        items = fields.Nested(
            lambda: model_validator.validator_class("array-items", strict=False)()
        )
        uniqueItems = fields.Boolean(required=False)
        minItems = fields.Integer(required=False)
        maxItems = fields.Integer(required=False)



import marshmallow as ma
from marshmallow import fields

from .utils import ExtendablePartSchema, PermissiveSchema


class PropertyJSONSchema(PermissiveSchema):
    generate = fields.Boolean(dump_default=True)



from collections import defaultdict
from functools import cached_property, partial
import importlib_metadata
import marshmallow as ma
from marshmallow.exceptions import ValidationError

PROPERTY_BY_TYPE_PREFIX = "property-by-type-"
ARRAY_ITEM_BY_TYPE_PREFIX = "array-item-by-type-"


class ModelValidator:
    @cached_property
    def validator_map(self):
        ret = {}
        for ep in importlib_metadata.entry_points().select(
            group="oarepo_model_builder.validation"
        ):
            validators = ep.load()
            for validator_type, validator in validators.items():
                ret.setdefault(
                    validator_type, []
                )  # create always, regardless if the validator is empty or not - to register empty extension points
                if isinstance(validator, (list, tuple)):
                    ret[validator_type].extend(validator)
                else:
                    ret[validator_type].append(validator)
        return ret

    def validator_class(self, section="root", strict=True):
        if section.startswith(PROPERTY_BY_TYPE_PREFIX):
            validators = self.get_property_validator_class(section)
        elif section.startswith(ARRAY_ITEM_BY_TYPE_PREFIX):
            validators = self.get_array_item_validator_class(section)
        else:
            validators = tuple(
                self.validator_map[section],
            )

        if not validators:
            raise ValidationError(f'Do not have validators for "{section}"')

        # if any of the validators is extremely permissive, do not add strict meta
        for v in validators:
            if hasattr(v, "Meta"):
                if getattr(v.Meta, "unknown", None) == ma.INCLUDE:
                    strict = False
                    break
        if strict:

            class Meta:
                unknown = ma.RAISE

            return type(f"{section.title()}Validator", validators, {"Meta": Meta})
        else:
            return type(f"{section.title()}Validator", validators, {})

    def get_property_validator_class(self, section):
        datatype_name = section[len(PROPERTY_BY_TYPE_PREFIX) :]
        return self._get_datatype_validator_class(
            section, datatype_name, "object-property"
        )

    def get_array_item_validator_class(self, section):
        datatype_name = section[len(ARRAY_ITEM_BY_TYPE_PREFIX) :]
        return self._get_datatype_validator_class(section, datatype_name, "array-item")

    def _get_datatype_validator_class(
        self, section, datatype_name, extra_validation_key
    ):
        from oarepo_model_builder.datatypes import datatypes

        validators = self.validator_map.get(section, ())

        datatype_class = datatypes.get_datatype_class(datatype_name)
        if not datatype_class:
            raise ValidationError(
                f"No datatype class registered for datatype {datatype_name}"
            )
        validators = tuple(
            [
                *validators,
                *self.validator_map["property"],
                *self.validator_map[extra_validation_key],
                datatype_class.ModelSchema,
            ]
        )
        return validators

    def validate(self, data):
        validator_class = self.validator_class()
        validator = validator_class()
        return validator.dump(validator.load(data))


model_validator = ModelValidator()



import marshmallow as ma
from marshmallow import fields
from marshmallow.exceptions import ValidationError
import typing
from functools import cached_property
import re
import typing
import marshmallow as ma
from marshmallow import fields
from marshmallow.error_store import ErrorStore
from marshmallow.exceptions import ValidationError


class CheckedConstant(fields.Constant):
    def _deserialize(self, value, *args, **kwargs):
        if value != self.constant:
            raise ValidationError(f"Bad value: '{self.constant}' expected.")
        return super()._deserialize(value, *args, **kwargs)


class ExtendablePartSchema(ma.Schema):
    extend = fields.String(required=False)
    use = fields.String(required=False)


class RegexOpts(ma.SchemaOpts):
    def __init__(self, meta, **kwargs):
        super().__init__(meta, **kwargs)
        self.regex_fields = getattr(meta, "regex_fields", [])


class RegexFieldsSchema(ma.Schema):
    OPTIONS_CLASS = RegexOpts

    def _deserialize(
        self,
        data: typing.Union[
            typing.Mapping[str, typing.Any],
            typing.Iterable[typing.Mapping[str, typing.Any]],
        ],
        *,
        error_store: ErrorStore,
        many: bool = False,
        partial=False,
        unknown=ma.RAISE,
        index=None,
    ):
        loaded = super()._deserialize(
            data,
            error_store=error_store,
            many=many,
            partial=partial,
            unknown=ma.EXCLUDE,  # exclude unknown - will be handled by the regex
            index=index,
        )
        # load all regex fields
        if many:
            for (loaded_data, orig_data) in zip(loaded, data):
                self.load_regex_fields(loaded_data, orig_data, error_store, index)
        else:
            self.load_regex_fields(loaded, data, error_store, index)
        return loaded

    @cached_property
    def compiled_regex_fields(self):
        return [(re.compile(x["key"]), x["field"]) for x in self.opts.regex_fields]

    def load_regex_fields(self, loaded_data, data, error_store, index):
        index_errors = self.opts.index_errors

        for fld in data:
            if fld in loaded_data:
                continue

            for fld_regex, field in self.compiled_regex_fields:
                if fld_regex.match(fld):
                    field = field()
                    self._bind_field(fld, field)
                    try:
                        loaded_data[fld] = field.deserialize(data[fld], fld, data)
                    except ValidationError as error:
                        error_store.store_error(error.messages, fld, index=index)
                    break

        data_keys = set(data.keys())
        loaded_keys = set(loaded_data.keys())
        for key in data_keys - loaded_keys:
            error_store.store_error(
                [self.error_messages["unknown"]],
                key,
                (index if index_errors else None),
            )

    def _serialize(self, obj, *, many: bool = False):
        serialized = super()._serialize(obj, many=many)
        if many:
            for (serialized_data, orig_data) in zip(serialized, obj):
                self.save_regex_fields(serialized_data, orig_data)
        else:
            self.save_regex_fields(serialized, obj)
        return serialized

    def save_regex_fields(self, serialized, obj):
        for attr_name in obj:
            if attr_name in serialized:
                continue

            for fld_regex, field in self.compiled_regex_fields:
                if fld_regex.match(attr_name):
                    self.save_regex_field(serialized, attr_name, field, obj)
                    break

    def save_regex_field(self, serialized, attr_name, field, obj):
        field = field()
        self._bind_field(attr_name, field)

        value = field.serialize(attr_name, obj, accessor=self.get_attribute)
        if value is ma.missing:
            return
        key = field.data_key if field.data_key is not None else attr_name
        serialized[key] = value


class PermissiveSchema(ExtendablePartSchema):
    class Meta:
        unknown = ma.INCLUDE

    def dump(self, obj, *, many=None):
        if not obj:
            return super().dump(obj, many=many)
        if many:
            return [self.dump(x, many=False) for x in obj]
        return {**obj, **super().dump(obj, many=False)}



from oarepo_model_builder.utils.jinja import package_name

from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioRecordMetadataModelsSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_metadata_models_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        metadata_package = package_name(self.current_model.record_metadata_class)

        output.add_entry_point(
            "invenio_db.models",
            self.current_model.record_schema_metadata_setup_cfg,
            metadata_package,
        )



from pathlib import Path

from oarepo_model_builder.datatypes import DataType, datatypes

from ..outputs.cfg import CFGOutput
from . import OutputBuilder, process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules


class ModelSaverBuilder(JSONBaseBuilder):
    TYPE = "model_saver"
    output_file_type = "json"
    output_file_name = "saved-model-file"
    parent_module_root_name = "models"

    @process("**")
    def model_element(self):
        self.model_element_enter()
        self.build_children()
        self.model_element_leave()

    def enter_model(self):
        # remove the special json base builder functionality
        # TODO: better handling for this
        pass

    def model_element_enter(self):
        super().model_element_enter()
        if self.stack.top.schema_element_type in ("items", "property"):
            datatype: DataType = datatypes.get_datatype(
                self.stack.top.data,
                self.stack.top.key,
                self.current_model,
                self.whole_schema,
            )
            if datatype:
                marshmallow = datatype.marshmallow()
                marshmallow.setdefault("imports", []).extend(
                    [
                        {"import": imp.import_path, "alias": imp.alias}
                        if imp.alias
                        else {
                            "import": imp.import_path,
                        }
                        for imp in datatype.imports()
                    ]
                )
                self.output.merge(
                    {
                        "marshmallow": marshmallow,
                    }
                )

    def output_primitive(self, top, data):
        if data is not None:
            if not isinstance(data, (str, float, int, bool)):
                data = str(data)
        return super().output_primitive(top, data)

    def begin(self, schema, settings):
        super().begin(schema, settings)

        output_name = self.current_model[self.output_file_name]
        self.output = self.builder.get_output(self.output_file_type, output_name)
        self.output.enter(self.whole_schema.model_field, {"type": "object"})
        ensure_parent_modules(
            self.builder, Path(output_name), ends_at=self.parent_module_root_name
        )

    def finish(self):
        self.output.leave()
        # force clean output
        super().finish()
        self.output.force_clean_output()


class ModelRegistrationBuilder(OutputBuilder):
    TYPE = "model_registration"

    def finish(self):
        super().finish()
        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")
        output.add_entry_point(
            "oarepo.models",
            self.current_model.oarepo_models_setup_cfg,
            f"{self.current_model.package}.models:{Path(self.current_model.saved_model_file).name}",
        )



import marshmallow as ma
from marshmallow import fields
from .model_validation import model_validator

from .utils import ExtendablePartSchema, CheckedConstant, PermissiveSchema


class ModelSchema(ExtendablePartSchema):
    type = CheckedConstant("object", required=False)
    properties = fields.Nested(
        lambda: model_validator.validator_class("properties", strict=False)()
    )
    opensearch = fields.Nested(
        lambda: model_validator.validator_class("model-opensearch", strict=False)()
    )
    plugins = (
        fields.Nested(
            lambda: model_validator.validator_class("plugins-schema")(), required=False
        ),
    )

    @ma.pre_load(pass_many=False)
    def set_properties_before_load(self, data, **kwargs):
        if not data.get("properties"):
            data = {**data}
            data["properties"] = {}
        return data


class ModelOpenSearchSchema(PermissiveSchema):
    pass



import os
from pathlib import Path
from typing import Dict
from oarepo_model_builder.utils.camelcase import camel_case, snake_case
from oarepo_model_builder.utils.deepmerge import deepmerge

from oarepo_model_builder.utils.jinja import split_base_name

from ..schema import ModelSchema
from . import ModelPreprocessor


class DefaultValuesModelPreprocessor(ModelPreprocessor):
    TYPE = "default"

    def transform(self, schema: ModelSchema, settings: Dict):
        model = schema.current_model

        deepmerge(
            model,
            {
                "record-prefix": camel_case(split_base_name(model.package)),
            },
        )

        self.set(
            model,
            "package",
            lambda: os.path.basename(
                schema.schema.get("output-directory", os.getcwd())
            ).replace("-", "_"),
        )

        self.set(
            model,
            "package-base",
            lambda: split_base_name(model.package),
        )

        self.set(model, "package-base-upper", lambda: model.package_base.upper())

        self.set(model, "kebap-package", lambda: model.package.replace("_", "-"))

        @self.set(model, "package-path")
        def c():
            package_path = model.package.split(".")
            return Path(package_path[0]).joinpath(*package_path[1:])

        self.set(model, "schema-version", lambda: schema.schema.get("version", "1.0.0"))

        self.set(
            model,
            "schema-name",
            lambda: f"{snake_case(model.record_prefix)}-{model.schema_version}.json",
        )

        self.set(
            model,
            "schema-file",
            lambda: os.path.join(
                model.package_path, "records", "jsonschemas", model.schema_name
            ),
        )

        self.set(model, "mapping-package", lambda: f"{model.package}.records.mappings")

        self.set(
            model,
            "jsonschemas-package",
            lambda: f"{model.package}.records.jsonschemas",
        )

        self.set(
            model,
            "mapping-file",
            lambda: os.path.join(
                model.package_path,
                "records",
                "mappings",
                "os-v2",
                snake_case(model.record_prefix),
                model.schema_name,
            ),
        )

        self.set(model, "schema-server", lambda: "http://localhost/schemas/")

        self.set(
            model,
            "index-name",
            lambda: snake_case(model.record_prefix)
            + "-"
            + os.path.basename(model.mapping_file).replace(".json", ""),
        )

        self.set(model, "collection-url", lambda: f"/{model.kebap_package}/")
        self.set(model, "model-name", lambda: model.package_base)

        # for outputting the model
        self.set(
            model,
            "saved-model-file",
            lambda: model.package_path / "models" / "model.json",
        )

        self.set(
            model,
            "oarepo-models-setup-cfg",
            lambda: snake_case(model.record_prefix),
        )



import copy
from typing import List

from oarepo_model_builder.utils.deepmerge import deepmerge


class JSONStackException(Exception):
    pass


class JSONStack:
    """Hierarchic json builder."""

    IGNORED_SUBTREE = object()
    IGNORED_NODE = object()

    def __init__(self):
        self.stack: List = [{}]

    def should_ignore(self, element):
        return element is self.IGNORED_SUBTREE or element is self.IGNORED_NODE

    def push(self, key, el):
        try:
            if key is None:
                assert isinstance(el, dict)
                assert self.empty
                self.stack[0] = deepmerge(copy.deepcopy(el), self.stack[0], [])
                return
            top = self.stack[-1]
            if top is self.IGNORED_SUBTREE:
                self.stack.append(self.IGNORED_SUBTREE)
            elif self.should_ignore(el):
                self.stack.append(el)
            else:
                if top is self.IGNORED_NODE:
                    top = self.real_top

                el = copy.deepcopy(el)
                if isinstance(top, dict):
                    if key not in top:
                        top[key] = el
                    else:
                        top[key] = deepmerge(el, top[key])
                elif isinstance(top, list):
                    if key < len(top):
                        top[key] = deepmerge(el, top[key])
                    else:
                        assert key == len(top)
                        top.append(el)
                else:
                    raise NotImplemented(
                        f"Set for datatype {type(top)} is not implemented"
                    )
                self.stack.append(el)
        except Exception as e:
            raise JSONStackException(
                f'Error pushing to json stack. Key "{key}", stack top {self.stack[-1]}'
            ) from e

    def pop(self):
        if not self.empty:
            self.stack.pop()

    @property
    def empty(self):
        return len(self.stack) == 1

    @property
    def value(self):
        return self.stack[0]

    @property
    def real_top(self):
        for t in reversed(self.stack):
            if t is not self.IGNORED_NODE:
                return t

    def merge(self, value):
        real_top = self.real_top
        real_top.update(deepmerge(value, real_top, []))



from pathlib import Path

from . import process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules


class JSONSchemaBuilder(JSONBaseBuilder):
    TYPE = "jsonschema"
    output_file_type = "jsonschema"
    output_file_name = "schema-file"
    parent_module_root_name = "jsonschemas"

    @process("**", condition=lambda current, stack: stack.schema_valid)
    def model_element(self):
        if isinstance(self.stack.top.data, dict) and not self.stack.top.data.get(
            "jsonschema", {}
        ).get("generate", True):
            # do not build the element if it should not be generated
            return
        self.model_element_enter()
        self.build_children()
        if (
            self.stack.top.schema_element_type == "items"
            or self.stack.top.schema_element_type == "property"
        ):
            self.check_and_output_required()
        self.merge_jsonschema(self.stack.top.data)
        self.model_element_leave()

    def merge_jsonschema(self, data):
        if isinstance(data, dict) and "jsonschema" in data:
            jsonschema = self.call_components(
                "before_merge_jsonschema", data["jsonschema"], stack=self.stack
            )
            self.output.merge_jsonschema(jsonschema)

    def check_and_output_required(self):
        top_data = self.stack.top.data
        if isinstance(top_data, dict) and "properties" in top_data:
            self.output.collect_required()

    def on_enter_model(self, output_name):
        self.output.primitive("type", "object")
        ensure_parent_modules(
            self.builder, Path(output_name), ends_at=self.parent_module_root_name
        )

    def on_leave_model(self):
        self.check_and_output_required()



from marshmallow import fields
from .model_validation import model_validator

from .utils import ExtendablePartSchema


class SettingsSchema(ExtendablePartSchema):
    opensearch = fields.Nested(
        lambda: model_validator.validator_class("settings-opensearch")()
    )

    python = fields.Nested(lambda: model_validator.validator_class("settings-python")())


class SettingsOpenSearchSchema(ExtendablePartSchema):
    version = fields.String(dump_default="os-v2")


class SettingsPythonSchema(ExtendablePartSchema):
    use_black = fields.Boolean(data_key="use-black", dump_default=True)
    use_isort = fields.Boolean(data_key="use-isort", dump_default=True)
    always_defined_import_prefixes = fields.List(
        fields.String(), data_key="always-defined-import-prefixes", required=False
    )
    templates = fields.Dict(fields.String(), fields.String())



from typing import List

from .datatypes import DataType, Import


class BaseDateDataType(DataType):
    marshmallow_field = "ma_fields.String"


class DateDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    model_type = "date"

    def mapping(self, **extras):
        return super().mapping(format="strict_date", **extras)

    def json_schema(self, **extras):
        return super().json_schema(format="date", **extras)

    def marshmallow_validators(self) -> List[str]:
        return super().marshmallow_validators() + [
            "validate_date('%Y:%m:%d')",
        ]

    def imports(self, *extra) -> List[Import]:
        return super().imports(
            Import(import_path="oarepo_runtime.validation.validate_date", alias=None)
        )


class TimeDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    model_type = "time"

    def mapping(self, **extras):
        return super().mapping(format="strict_time||strict_time_no_millis")

    def json_schema(self, **extras):
        return super().json_schema(format="time", **extras)

    def marshmallow_validators(self) -> List[str]:
        return super().marshmallow_validators() + [
            "validate_date('%H:%M:%S')",
        ]

    def imports(self, *extra) -> List[Import]:
        return super().imports(
            Import(import_path="oarepo_runtime.validation.validate_date", alias=None)
        )


class DateTimeDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    marshmallow_field = "mu_fields.ISODateString"
    model_type = "datetime"

    def mapping(self, **extras):
        return super().mapping(format="strict_date_time||strict_date_time_no_millis")

    def json_schema(self, **extras):
        return super().json_schema(format="date-time", **extras)


class EDTFDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    model_type = "edtf"

    def mapping(self, **extras):
        return super().mapping(
            format="strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy"
        )

    def marshmallow_validators(self):
        return super().marshmallow_validators() + [
            "mu_fields.EDTFValidator(types=EDTFDate)"
        ]

    def imports(self, *args):
        return super().imports(Import(name="edtf.Date", alias="EDTFDate"))


class EDTFIntervalType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date_range"
    model_type = "edtf-interval"

    def mapping(self, **extras):
        return super().mapping(
            format="strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy"
        )

    def marshmallow_validators(self):
        return super().marshmallow_validators() + [
            "mu_fields.EDTFValidator(types=EDTFInterval)"
        ]

    def imports(self, *args):
        return super().imports(Import(name="edtf.Interval", alias="EDTFInterval"))



from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.outputs.python import PythonOutput


class InvenioVersionBuilder(PythonBuilder):
    TYPE = "invenio_version"

    def finish(self):
        super().finish()

        python_output: PythonOutput = self.builder.get_output(
            "python", self.current_model.package_path / "version.py"
        )
        python_output.merge(
            "invenio_version", {"settings": self.settings, "current_model": self.current_model}
        )



from .invenio_base import InvenioBaseClassPythonBuilder
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch
from ..utils.jinja import base_name


class InvenioRecordServiceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_service_config"
    class_config = "record-service-config-class"
    template = "record-service-config"

    def process_template(self, python_path, template, **extra_kwargs):
        if self.parent_modules:
            self.create_parent_modules(python_path)
        output: PythonOutput = self.builder.get_output("python", python_path)

        context = HyphenMunch(
            settings=self.settings,
            current_model=self.current_model,
            **extra_kwargs
        )
        template = self.call_components(
            "invenio_before_python_template", template, context=context
        )
        output.merge(template, context)



from marshmallow import fields

from .utils import ExtendablePartSchema


class PropertyFacets(ExtendablePartSchema):
    clazz = fields.String(data_key="class", required=False)
    field = fields.String(required=False)



import marshmallow as ma
from marshmallow import fields
from .model_validation import model_validator
from marshmallow_union import Union

from .utils import CheckedConstant


class PluginsSchema(ma.Schema):
    packages = fields.List(fields.String())
    output = fields.Nested(
        lambda: model_validator.validator_class("plugin-schema")(), required=False
    )
    builder = fields.Nested(
        lambda: model_validator.validator_class("plugin-schema")(), required=False
    )
    model = fields.Nested(
        lambda: model_validator.validator_class("plugin-schema")(), required=False
    )
    property = fields.Nested(
        lambda: model_validator.validator_class("plugin-schema")(), required=False
    )


class PluginConfigSchema(ma.Schema):
    disable = Union(CheckedConstant("__all__"), fields.List(fields.String()))
    enable = fields.List(fields.String())
    include = fields.List(fields.String())
